"""Testing opening of datasets."""

import importlib
import os
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

import fsspec
import pytest
import xarray as xr
from xarray.testing import assert_identical

import ecmwfspec
from ecmwfspec import xr_accessor  # noqa: F401


def test_xr_accessor(patch_dir: Path, zarr_file: Path) -> None:
    """Test staging."""
    zarr_file1 = [*zarr_file.rglob("*.zarr")][0]
    urls = [f"ec://{zarr_file1}"]

    dataset = xr.open_dataset(
        urls[0],
        engine="zarr",
        chunks={"time": 1, "x": 100, "y": 100},
        backend_kwargs={"storage_options": {"ec": {"ec_cache": patch_dir}}},
    )
    dataset_original = xr.open_dataset(zarr_file1, engine="zarr")
    path_to_precip_chunks = os.path.join(
        patch_dir, urls[0].replace("ec:///", ""), "precip"
    )
    assert dataset["precip"]._in_memory is False, "dataset has been loaded into memory"
    assert (
        os.path.exists(path_to_precip_chunks) is False
    ), "zarr directory exists, though chunks should not exist yet"
    dataset["precip"].ecfs.stage()
    assert dataset["precip"]._in_memory is False, "dataset has been loaded into memory"
    assert os.path.exists(path_to_precip_chunks), "chunks were not requested"
    files = os.listdir(path_to_precip_chunks)
    assert len(files) == 2, "chunks files are missing"
    dataset_original.load()
    shutil.rmtree(zarr_file1)
    dataset["precip"].load()
    assert_identical(dataset.precip, dataset_original.precip)


def test_reading_dataset(patch_dir: Path, netcdf_files: Path) -> None:
    """Test reading datafiles with ecmwfspec."""
    import fsspec

    urls = [
        fsspec.open(f"ec://{f}", ec_cache=patch_dir, override=True).open()
        for f in netcdf_files.rglob("*.nc")
    ]
    dataset1 = xr.open_mfdataset(list(netcdf_files.rglob("*.nc")), combine="by_coords")
    dataset2 = xr.open_mfdataset(urls, combine="by_coords")
    assert dataset1 == dataset2


@mock.patch.dict(os.environ, {}, clear=True)
def test_errors(patch_dir: Path) -> None:
    """Check if ec specs warns the users if the cache wasn't set and fallback
    not available."""
    importlib.reload(ecmwfspec)
    importlib.reload(fsspec)

    with pytest.raises(ValueError):
        fsspec.open("ec:///foo/bar.txt", mode="rt").open()

    ecmwfspec.core.FileQueue.queue.clear()  # TODO: empty queue automatically


@mock.patch.dict(os.environ, {"SCRATCH": str(TemporaryDirectory())}, clear=True)
def test_warnings(patch_dir: Path) -> None:
    """Check if ec specs warns the users if the cache wasn't set and fallback
    not available."""
    importlib.reload(fsspec)

    with pytest.warns(UserWarning):
        fsspec.open("ec:///foo/bar.txt", mode="rt").open()

    ecmwfspec.core.FileQueue.queue.clear()  # TODO: empty queue automatically


def test_reading_nonexisting_dataset(patch_dir: Path, netcdf_files: Path) -> None:
    """Test read-failure on non-existing files."""
    import fsspec

    non_existing_urls = fsspec.open(
        "ec:///foo/bar.nc",
        ec_cache=patch_dir,
        mode="rt",
    ).open()
    with pytest.raises(FileNotFoundError):
        non_existing_urls.read()


def test_text_mode(patch_dir: Path) -> None:
    """Test opening the files in text mode."""
    import fsspec

    with TemporaryDirectory() as temp_dir:
        inp_file = Path(temp_dir) / "foo.txt"
        write_file = patch_dir.joinpath(*inp_file.parts[1:])
        write_file.parent.mkdir(exist_ok=True, parents=True)
        print(write_file)
        with write_file.open("w") as f_obj:
            f_obj.write("foo")
        url = fsspec.open(
            f"ec:///{inp_file}",
            ec_cache=patch_dir,
            override=False,
            mode="rt",
        ).open()
    assert Path(url.name) == write_file
    assert url.tell() == 0
    assert url.read() == "foo"


def test_ro_mode(patch_dir: Path) -> None:
    """Check if ec specs is ro."""
    import fsspec

    with pytest.raises(NotImplementedError):
        fsspec.open("ec:///foo/bar.nc", mode="w").open()

    url = fsspec.open("ec:///foo/bar.txt", ec_cache="foo").open()

    with pytest.raises(NotImplementedError):
        url.writelines()

    with pytest.raises(NotImplementedError):
        url.write()

    assert url.writable() is False


def test_list_files(patch_dir: Path, netcdf_files: Path) -> None:
    """Test listing the files."""
    import fsspec

    files = list(netcdf_files.iterdir())
    ec = fsspec.filesystem("ec", ec_cache=patch_dir)
    res = ec.ls(netcdf_files, detail=False)
    assert len(files) == len(res)
    res = ec.ls(netcdf_files, detail=True)
    for info in res:
        assert isinstance(info, dict)
        assert "name" in info
        assert "type" in info
        assert "size" in info
        assert info["size"] is None



from __future__ import annotations

import io
import logging
import os
import threading
import time
import warnings
from pathlib import Path
from queue import Queue
from typing import (
    IO,
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    TypedDict,
    Union,
    overload,
)

from fsspec.spec import AbstractFileSystem

from . import ecfs_wrapper as ecfs

logger = logging.getLogger("ecmwfspec")
logger.setLevel(logging.DEBUG)


MAX_RETRIES = 2
FileQueue: Queue[Tuple[str, str]] = Queue(maxsize=-1)
FileInfo = TypedDict(
    "FileInfo", {"name": str, "size": Literal[None], "type": Union[str | None]}
)
_retrieval_lock = threading.Lock()


class ECFile(io.IOBase):
    """File handle for files on the hsm archive.

    Parameters
    ----------
    url: str
        Source path of the file that should be retrieved.
    local_file: str
        Destination path of the downloaded file.
    ec_cache: str | Path
        Destination of the temporary storage. This directory is used to
        retrieve data from tape.
    override: bool, default: False
        Override existing files
    touch: bool, default: True
        Update existing files on the temporary storage to prevent them
        from being deleted.
    mode: str, default: rb
        Specify the mode in which the files are opened

        'r'       open for reading (default)
        'b'       binary mode (default)
        't'       text mode
    file_permissions: int, default 0o3777
        Permission when creating directories and files.
    **kwargs:
        Additional keyword arguments passed to the open file descriptor method.

    Example
    -------

    Use fsspec to open data stored on tape, temporary data will be downloaded
    to a central scratch folder:

    ::

        import ffspec
        import xarray as xr

        url = fsspec.open("ec:////arch/bb1203/data.nc",
                          ec_cache="/scratch/b/b12346").open()
        dset = xr.open_dataset(url)
    """

    write_msg: str = "Write mode is not suppored"
    """Error message that is thrown if the files are attempted to be opened in
    any kind of write mode."""

    def __init__(
        self,
        url: str,
        local_file: str,
        ec_cache: Union[str, Path],
        *,
        override: bool = True,
        mode: str = "rb",
        touch: bool = True,
        file_permissions: int = 0o3777,
        delay: int = 2,
        _lock: threading.Lock = _retrieval_lock,
        _file_queue: Queue[Tuple[str, str]] = FileQueue,
        **kwargs: Any,
    ):
        if not set(mode) & set("r"):  # The mode must have a r
            raise NotImplementedError(self.write_msg)
        if "b" not in mode:
            kwargs.setdefault("encoding", "utf-8")
        self._file = str(Path(local_file).expanduser().absolute())
        self._url = str(url)
        self.eccache = Path(ec_cache)
        self.touch = touch
        self.file_permissions = file_permissions
        self._order_num = 0
        self._file_obj: Optional[IO[Any]] = None
        self._lock = _lock
        self.kwargs = kwargs
        self.mode = mode
        self.newlines = None
        self.error = "strict"
        self.encoding = kwargs.get("encoding")
        self.write_through = False
        self.delay = delay
        self._file_queue = _file_queue
        print(self._file)
        with _lock:
            if not Path(self._file).exists() or override:
                self._file_queue.put((self._url, str(Path(self._file).parent)))
            elif Path(self._file).exists():
                if self.touch:
                    Path(self._file).touch()
                self._file_obj = open(self._file, mode, **kwargs)

    @property
    def name(self) -> str:
        """Get the file for the ECFile object."""
        if self._file_obj is not None:
            return self._file
        return self._url

    def _retrieve_items(self, retrieve_files: list[tuple[str, str]]) -> None:
        """Get items from the tape archive."""

        retrieval_requests: List[str] = list()
        logger.debug("Retrieving %i items from ECFS", len(retrieve_files))
        for inp_file, _ in retrieve_files:
            retrieval_requests.append(inp_file)
        for file in retrieval_requests:
            logger.debug("Retrieving file: %s", file)
            local_path = self.eccache / Path(file.strip("/"))
            ecfs.cp("ec:" + file, str(local_path))
            local_path.chmod(self.file_permissions)

    def _cache_files(self) -> None:
        time.sleep(self.delay)
        with self._lock:
            items = []
            if self._file_queue.qsize() > 0:
                self._file_queue.put(("finish", "finish"))
                for _ in range(self._file_queue.qsize() - 1):
                    items.append(self._file_queue.get())
                    self._file_queue.task_done()
                try:
                    self._retrieve_items(items)
                except Exception as error:
                    _ = [
                        self._file_queue.get() for _ in range(self._file_queue.qsize())
                    ]
                    self._file_queue.task_done()
                    raise error
                _ = self._file_queue.get()
                self._file_queue.task_done()
        self._file_queue.join()
        self._file_obj = open(self._file, self.mode, **self.kwargs)

    def __fspath__(self) -> str:
        if self._file_obj is None:
            self._cache_files()
        return self.name

    def tell(self) -> int:
        if self._file_obj is None:
            self._cache_files()
        return self._file_obj.tell()  # type: ignore

    def seek(self, target: int, whence: int = 0) -> int:  # type: ignore
        if self._file_obj is None:
            self._cache_files()
        return self._file_obj.seek(target, whence)  # type: ignore

    @staticmethod
    def readable() -> Literal[True]:
        """Compatibility method."""
        return True

    @staticmethod
    def writeable() -> Literal[False]:
        """Compatibility method."""
        return False

    @staticmethod
    def seekable() -> Literal[True]:
        """Compatibility method."""
        return True

    def read(self, size: int = -1) -> str:
        """The the content of a file-stream.

        size: int, default: -1
            read at most size characters from the stream, -1 means everything
            is read.
        """
        if self._file_obj is None:
            self._cache_files()
        return self._file_obj.read(size)  # type: ignore

    @staticmethod
    def flush() -> None:
        """Flushing file systems shouldn't work for ro modes."""
        return None

    def writelines(self, *arg: Any) -> None:
        """Compatibility method."""
        raise NotImplementedError(self.write_msg)

    def write(self, *arg: Any) -> None:
        """Writing to tape is not supported."""
        raise NotImplementedError(self.write_msg)

    def close(self) -> None:
        if self._file_obj is not None:
            self._file_obj.close()


class ECFileSystem(AbstractFileSystem):
    """Abstract class for hsm files systems.

    The implementation intracts with the hsm tape storage system, files
    that are accessed are downloaded to a temporary data storage.

    Parameters
    ----------

    ec_cache: str | Path, default: None
        Destination of the temporary storage. This directory is used to
        retrieve data from tape.
    block_size: int, default: None
         Some indication of buffering - this is a value in bytes
    file_permissions: int, default: 0o3777
        Permission when creating directories and files.
    override: bool, default: False
        Override existing files
    touch: bool, default: True
        Update existing files on the temporary storage to prevent them
        from being deleted.
    **storage_options:
        Additional options passed to the AbstractFileSystem class.
    """

    protocol = "ec"
    local_file = True
    sep = "/"

    def __init__(
        self,
        block_size: Optional[int] = None,
        ec_cache: Optional[Union[str, Path]] = None,
        file_permissions: int = 0o3777,
        touch: bool = True,
        delay: int = 2,
        override: bool = False,
        **storage_options: Any,
    ):
        super().__init__(
            block_size=block_size,
            asynchronous=False,
            loop=None,
            **storage_options,
        )
        ec_options = storage_options.get("ec", {})
        ec_cache = (
            ec_options.get("ec_cache", None) or ec_cache or os.environ.get("EC_CACHE")
        )
        if not ec_cache:
            ec_cache = os.environ.get("SCRATCH", None)
            if ec_cache is None:
                raise ValueError("No cache directory specified.")
            else:
                warnings.warn(
                    "Neither the ec_cache argument nor the EC_CACHE environment "
                    "variable is set. Falling back to default "
                    f"{ec_cache}",
                    UserWarning,
                    stacklevel=2,
                )
        self.touch = touch
        self.ec_cache = Path(ec_cache)
        self.override = override
        self.delay = delay
        self.file_permissions = file_permissions

    @overload
    def ls(
        self, path: Union[str, Path], detail: Literal[True], **kwargs: Any
    ) -> List[FileInfo]: ...

    @overload
    def ls(
        self, path: Union[str, Path], detail: Literal[False], **kwargs: Any
    ) -> List[str]: ...

    def ls(
        self, path: Union[str, Path], detail: bool = True, **kwargs: Any
    ) -> Union[List[FileInfo], List[str]]:
        """List objects at path.

        This includes sub directories and files at that location.

        Parameters
        ----------
        path: str | pathlib.Path
            Path of the file object that is listed.
        detail: bool, default: True
            if True, gives a list of dictionaries, where each is the same as
            the result of ``info(path)``. If False, gives a list of paths
            (str).


        Returns
        -------
        list : List of strings if detail is False, or list of directory
               information dicts if detail is True.
        """
        path = Path(path)
        filelist = ecfs.ls(str(path), detail=detail)
        # Drop summary line of detailed listing
        if detail:
            filelist = filelist[filelist.permissions != "total"]
        detail_list: List[FileInfo] = []
        types = {"d": "directory", "-": "file"}
        detail_list = [
            {
                "name": str(path / file_entry.path),
                "size": None,  # sizes are human readable not in bytes
                "type": types[file_entry.permissions[0]] if detail else None,
            }
            for _, file_entry in filelist.iterrows()
        ]
        if detail:
            return detail_list
        else:
            return [d["name"] for d in detail_list]

    def _open(
        self,
        path: str | Path,
        mode: str = "rb",
        block_size: Optional[int] = None,
        autocommit: bool = True,
        cache_options: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> ECFile:
        path = Path(self._strip_protocol(path))
        local_path = self.ec_cache.joinpath(*path.parts[1:])
        return ECFile(
            str(path),
            str(local_path),
            self.ec_cache,
            mode=mode,
            override=self.override,
            touch=self.touch,
            delay=self.delay,
            encoding=kwargs.get("encoding"),
            file_permissions=self.file_permissions,
        )



"""Wrapper of ECFS file system commands."""

import logging
import subprocess
from pathlib import Path
from typing import Union

import pandas as pd

logger = logging.getLogger(__name__)


def ls(
    path: Union[str, Path],
    detail: bool = False,
    allfiles: bool = False,
    recursive: bool = False,
    directory: bool = False,
) -> pd.DataFrame:
    """List files in a directory."""
    command = ["els", path]
    columns = ["path"]

    if detail:
        command.append("-l")
        columns = [
            "permissions",
            "links",
            "owner",
            "group",
            "size",
            "month",
            "day",
            "time",
            "path",
        ]

    if allfiles:
        command.append("-a")

    if directory:
        command.append("-d")

    if recursive:
        logger.warning(
            "Recursive option should be avoided on very large ECFS directory tress because of timeout issues."
        )
        command.append("-R")

    result = subprocess.run(
        command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True
    )
    logger.debug(result.stdout)

    if result.stderr is not None:
        logger.error(result.stderr)
        raise Exception("Error running command: {}".format(command))

    files = result.stdout.split("\n")  # type: ignore
    files = [f for f in files if f != ""]

    if detail:
        files = [f.split() for f in files]

    df = pd.DataFrame(files, columns=columns)

    return df


def cp(src: Union[str, Path], dst: Union[str, Path]) -> None:
    """Copy a file from src to dst."""
    command = ["ecp", src, dst]
    result = subprocess.check_output(command, text=True)
    logger.debug(result)

    if result != "":
        logger.error(result)
        raise Exception("Error running command: {}".format(command))

    return



"""Pytest definitions to run the unittests."""

from __future__ import annotations

import builtins
import os
import shutil
from pathlib import Path
from subprocess import PIPE, run
from tempfile import TemporaryDirectory
from typing import Generator, Union

import mock
import numpy as np
import pandas as pd
import pytest
import xarray as xr


class ECMock:
    """A mock that emulates what pyec is doing."""

    def __init__(self, _cache: dict[int, builtins.list[str]] = {}) -> None:
        self._cache = _cache

    def ec_list(self, inp_path: str) -> str:
        """Mock the ec_list method."""
        res = (
            run(["ls", "-l", inp_path], stdout=PIPE, stderr=PIPE)
            .stdout.decode()
            .split("\n")
        )
        return "\n".join(res[1:] + [res[0]])

    def ls(
        self,
        inp_path: Union[str, Path],
        detail: bool = False,
        allfiles: bool = False,
        recursive: bool = False,
        directory: bool = False,
    ) -> pd.DataFrame:
        """List files in a directory."""
        command = ["ls", inp_path]
        columns = ["path"]

        if detail:
            command.append("-l")
            columns = [
                "permissions",
                "links",
                "owner",
                "group",
                "size",
                "month",
                "day",
                "time",
                "path",
            ]

        if allfiles:
            command.append("-a")

        if directory:
            command.append("-d")

        if recursive:
            command.append("-R")

        result = run(command, stdout=PIPE, stderr=PIPE, text=True)

        files = result.stdout.split("\n")
        files = [f for f in files if f != ""]

        if detail:
            files_incl_details = [f.split() for f in files]
            df = pd.DataFrame(files_incl_details, columns=columns)
        else:
            df = pd.DataFrame(files, columns=columns)

        return df

    def cp(self, inp_path: str, out_path: str) -> None:
        """Mock the ecp method."""
        inp_path = inp_path.replace("ec:", "")
        os.makedirs(os.path.dirname(out_path), exist_ok=True)
        _ = (
            run(["cp", inp_path, out_path], stdout=PIPE, stderr=PIPE)
            .stdout.decode()
            .split("\n")
        )
        return

    def search(self, inp_f: builtins.list[str]) -> int | None:
        """Mock ec_search."""
        if not inp_f:
            return None
        hash_value = abs(hash(",".join(inp_f)))
        self._cache[hash_value] = inp_f
        return hash_value

    def ec_gen_file_query(self, inp_files: builtins.list[str]) -> builtins.list[str]:
        """Mock ec_gen_file_qeury."""
        return [f for f in inp_files if Path(f).exists()]

    def ec_retrieve(self, search_id: int, out_dir: str, preserve_path: bool) -> None:
        """Mock ec_retrieve."""
        for inp_file in map(Path, self._cache[search_id]):
            if preserve_path:
                outfile = Path(out_dir) / Path(str(inp_file).strip(inp_file.root))
                outfile.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy(inp_file, outfile)
            else:
                shutil.copy(inp_file, Path(out_dir) / inp_file.name)


def create_data(variable_name: str, size: int) -> xr.Dataset:
    """Create a xarray dataset."""
    coords: dict[str, np.ndarray] = {}
    coords["x"] = np.linspace(-10, -5, size)
    coords["y"] = np.linspace(120, 125, size)
    lat, lon = np.meshgrid(coords["y"], coords["x"])
    lon_vec = xr.DataArray(lon, name="Lg", coords=coords, dims=("y", "x"))
    lat_vec = xr.DataArray(lat, name="Lt", coords=coords, dims=("y", "x"))
    coords["time"] = np.array(
        [
            np.datetime64("2020-01-01T00:00"),
            np.datetime64("2020-01-01T12:00"),
            np.datetime64("2020-01-02T00:00"),
            np.datetime64("2020-01-02T12:00"),
        ]
    )
    dims = (4, size, size)
    data_array = np.empty(dims)
    for time in range(dims[0]):
        data_array[time] = np.zeros((size, size))
    dset = xr.DataArray(
        data_array,
        dims=("time", "y", "x"),
        coords=coords,
        name=variable_name,
    )
    data_array = np.zeros(dims)
    return xr.Dataset({variable_name: dset, "Lt": lon_vec, "Lg": lat_vec}).set_coords(
        list(coords.keys())
    )


@pytest.fixture(scope="session")
def patch_dir() -> Generator[Path, None, None]:
    with TemporaryDirectory() as temp_dir:
        with mock.patch("ecmwfspec.core.ecfs", ECMock()):
            yield Path(temp_dir)


@pytest.fixture(scope="session")
def save_dir() -> Generator[Path, None, None]:
    """Create a temporary directory."""
    with TemporaryDirectory() as td:
        yield Path(td)


@pytest.fixture(scope="session")
def data() -> Generator[xr.Dataset, None, None]:
    """Define a simple dataset with a blob in the middle."""
    dset = create_data("precip", 100)
    yield dset


@pytest.fixture(scope="session")
def netcdf_files(
    data: xr.Dataset,
) -> Generator[Path, None, None]:
    """Save data with a blob to file."""

    with TemporaryDirectory() as td:
        for time in (data.time[:2], data.time[2:]):
            time1 = pd.Timestamp(time.values[0]).strftime("%Y%m%d%H%M")
            time2 = pd.Timestamp(time.values[1]).strftime("%Y%m%d%H%M")
            out_file = (
                Path(td)
                / "the_project"
                / "test1"
                / "precip"
                / f"precip_{time1}-{time2}.nc"
            )
            out_file.parent.mkdir(exist_ok=True, parents=True)
            data.sel(time=time).to_netcdf(out_file)
        yield Path(td)


@pytest.fixture(scope="session")
def zarr_file(
    data: xr.Dataset,
) -> Generator[Path, None, None]:
    """Save data with a blob to file."""

    with TemporaryDirectory() as td:
        out_file = Path(td) / "the_project" / "test1" / "precip" / "precip.zarr"
        out_file.parent.mkdir(exist_ok=True, parents=True)
        data.to_zarr(out_file)
        yield Path(td)



from itertools import compress
from typing import Any, Union

import dask
import dask.array as da
import xarray as xr


@xr.register_dataarray_accessor("ecfs")
@xr.register_dataset_accessor("ecfs")
class ecfs:
    """Request files without keeping them in memory.

    This accessor is similar to :py:func:`xarray.Dataset.load`,
    :py:func:`xarray.Dataset.persist` and
    :py:meth:`xarray.Dataset.compute`, but instead of loading the
    dataset into memory or start the computation, it just requests the
    data and immediatly discards it. This allows the backend to trigger
    the retrieval of the data for later usage and can be cached locally
    when using e.g.`simplecache` or `ecmwfspec` with EC_CACHE.
    """

    def __init__(self, xarray_obj: Union[xr.Dataset, xr.DataArray]) -> None:
        self._obj = xarray_obj

    def stage(self) -> None:
        if isinstance(self._obj, xr.Dataset):
            self._get_dataset()
        elif isinstance(self._obj, xr.DataArray):
            self._get_dataarray()
        return

    def _check_layer(self, key: str, layer: str = "open_dataset") -> bool:
        """Check if dask graph key is part of `layer`

        Inputs
        ------
        key : str
            Dask task key
        layer : str
            Task layer identifier

        Returns
        -------
        result : bool
            Flag if key belongs to `layer`
        """
        return bool(key.startswith(layer))

    def _get_output_tasks(
        self, highLevelGraph: dask.highlevelgraph.HighLevelGraph
    ) -> list:
        """Get all tasks that are responsible for loading chunks, i.e. those in
        the `open_dataset` layers in the dask `highLevelGraph`.

        Inputs
        ------
        highLevelGraph : dask.highLevelGraph
            Entire dask task graph

        Returns
        -------
        result : list
            List of open_dataset tasks
        """
        k = list(highLevelGraph.keys())
        layer_mask = list(map(self._check_layer, highLevelGraph.keys()))

        return list(compress(k, layer_mask))

    def _get_output_keys(self, graph: dask.highlevelgraph.HighLevelGraph) -> set:
        layer_keys = self._get_output_tasks(graph.layers)
        layers = [graph.layers[k] for k in layer_keys]
        output_keys = [lay.get_output_keys() for lay in layers]

        return set().union(*output_keys)

    def _get_input_graph(
        self, graph: dask.highlevelgraph.HighLevelGraph, output_keys: set
    ) -> dask.highlevelgraph.HighLevelGraph:
        """Simplify dask graph based on `output_keys`"""
        input_graph = graph.cull(keys=output_keys)
        return input_graph

    def _connect_tasks(self, graph: dict, tasks_to_connect: set) -> dict:
        """Connect tasks with pseudo gathering the results of sub-tasks.

        Inputs
        ------
        graph : dict
            Dictionary representation of dask task graph
        tasks_to_connect : set
            List of tasks that should be connected embarrassingly parallel.

        Returns
        -------
        graph : dict
            Dask task graph with additional connections
        """

        def do_nothing(x: Any) -> None:
            return

        # Adding embarrassingly parallel layer to each task to immediatly
        # free memory
        for k, key in enumerate(tasks_to_connect):
            graph[f"do_nothing_w_dataset-{k}"] = (do_nothing, key)
        # Gather tasks
        k = len(tasks_to_connect)
        graph["do_nothing_at_all"] = (
            do_nothing,
            [f"do_nothing_w_dataset-{t}" for t in range(k)],
        )
        return graph

    def _get_data(self, data: da.Array) -> None:
        """Main function."""
        dask_keys = data.__dask_keys__()
        graph = data.dask.cull(keys=dask_keys)

        output_keys = self._get_output_keys(graph)
        input_graph = self._get_input_graph(graph, output_keys)
        input_graph = input_graph.to_dict()
        input_graph = self._connect_tasks(input_graph, output_keys)
        scheduler = (
            dask.base.get_scheduler()
        )  # determine whether LocalCluster/SLUMCluster etc. exist
        if scheduler is None:
            _ = dask.threaded.get(input_graph, "do_nothing_at_all")
        else:
            _ = scheduler(input_graph, "do_nothing_at_all")
        return

    def _get_dataarray(self) -> None:
        """Get_data wrapper for xr.DataArray."""
        self._get_data(self._obj.data)
        return

    def _get_dataset(self) -> None:
        """Get_data wrapper for xr.Dataset."""
        for var in self._obj.data_vars:
            self._get_data(self._obj[var].data)
        return
